/**
 * EasyBeans
 * Copyright (C) 2006-2011 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: FileUtils.java 5810 2011-03-17 13:31:23Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.file;

import java.io.BufferedReader;
import java.io.Closeable;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.Flushable;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Enumeration;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * This class manages operation done many times on files, like copying them.
 * @author Florent Benoit
 */
public final class FileUtils {

    /**
     * Size of the buffer.
     */
    private static final int BUFFER_SIZE = 4096;

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(FileUtils.class);

    /**
     * Date format.
     */
    private static final String DATE_FORMAT = "_yyyy.MM.dd-HH.mm.ss";

    /**
     * Utility class, no public constructor.
     */
    private FileUtils() {

    }

    /**
     * Unpack the source archive in a given directory.
     * @param packedJar source JarFile to be unpacked
     * @param dest the destination folder
     * @throws FileUtilsException When unpack fails
     */
    public static void unpack(final JarFile packedJar, final File dest) throws FileUtilsException {

        JarEntry entry = null;

        // get entries of the jar file
        Enumeration<JarEntry> entries = packedJar.entries();
        while (entries.hasMoreElements()) {
            entry = entries.nextElement();

            // File entry
            File entryFile = new File(dest, entry.getName());

            // Create directory
            if (entry.isDirectory()) {
                if (!entryFile.exists()) {
                    // create parent directories (with mkdirs)
                    if (!entryFile.mkdirs()) {
                        String err = "Can not create directory " + entryFile + ", Check the write access.";
                        throw new FileUtilsException(err);
                    }
                }
                continue;
            }

            // If it's a file, we must extract the file
            // Ensure that the directory exists.
            entryFile.getParentFile().mkdirs();

            InputStream is = null;
            // get the input stream
            try {
                is = packedJar.getInputStream(entry);
                // Dump to the file
                dump(is, entryFile);
            } catch (IOException ioe) {
                throw new FileUtilsException("Cannot get inputstream of entry '" + entry + "' of file '" + packedJar + "'.");
            } finally {
                try {
                    is.close();
                } catch (IOException ioe) {
                    logger.error("Cannot close input stream", ioe);
                }
            }

        }
    }

    /**
     * Write the given input stream in the given file.
     * @param in the inputStream to copy.
     * @param entryFile the file where the inputStream must be dumped.
     * @throws FileUtilsException if the dump failed.
     */
    public static void dump(final InputStream in, final File entryFile) throws FileUtilsException {

        try {
            // File output
            FileOutputStream out = new FileOutputStream(entryFile);
            int n = 0;
            try {
                // buffer
                byte[] buffer = new byte[BUFFER_SIZE];
                n = in.read(buffer);
                while (n > 0) {
                    out.write(buffer, 0, n);
                    n = in.read(buffer);
                }
                // Flush
                out.flush();
            } finally {
                out.close();
            }
        } catch (IOException e) {
            String err = "Error while unpacking entry " + entryFile + " : ";
            throw new FileUtilsException(err, e);
        }
    }

    /**
     * Print a text file in a print writer.
     * @param txtFile the text file
     * @param printWriter the print writer
     * @throws FileUtilsException if error occurs
     */
    public static void print(final File txtFile, final PrintWriter printWriter) throws FileUtilsException {
        try {
            BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(new FileInputStream(txtFile)));
            String line = bufferedReader.readLine();
            while (line != null) {
                printWriter.println(line);
                line = bufferedReader.readLine();
            }
            bufferedReader.close();
        } catch (IOException e) {
            throw new FileUtilsException("Error while printing", e);
        }
    }

    /**
     * Print a text file in standard output.
     * @param txtFile the text file
     * @throws FileUtilsException if error occurs
     */
    public static void print(final File txtFile) throws FileUtilsException {
        print(txtFile, new PrintWriter(System.out, true));
    }

    /**
     * Copy a file and return its copy.
     * @param src source file
     * @param dest dest file
     * @return the copied file
     * @throws FileUtilsException if the copy of the file failed
     */
    public static File copyFile(final String src, final String dest) throws FileUtilsException {
        return copyFile(new File(src), new File(dest));
    }

    /**
     * Copy a file and return its copy.
     * @param src source file
     * @param dest dest file
     * @return the copied file
     * @throws FileUtilsException if the copy of the file failed
     */
    public static File copyFile(final File src, final File dest) throws FileUtilsException {
        File newDest = null;
        if (dest.isDirectory()) {
            logger.debug("Copy a file to a directory, append source filename to directory.");
            newDest = new File(dest, src.getName());
        } else {
            newDest = dest;
        }

        FileInputStream fIn = null;
        FileOutputStream fOut = null;
        FileChannel sourceFC = null;
        FileChannel targetFC = null;
        try {

            // InputStream
            fIn = new FileInputStream(src);
            fOut = new FileOutputStream(newDest);

            // nio channel
            sourceFC = fIn.getChannel();
            targetFC = fOut.getChannel();

            // transfer
            targetFC.transferFrom(sourceFC, 0, sourceFC.size());

            // ensure all has been written
            targetFC.force(true);
        } catch (Exception e) {
            throw new FileUtilsException("Error during copy file : " + src + " -> " + newDest, e);
        } finally {
            close(sourceFC, targetFC);
            flush(fOut);
            close(fOut, fIn);
        }

        return newDest;
    }

    /**
     * @param path file/directory to be deleted
     * @return true if deletion was OK
     */
    public static boolean delete(final String path) {
        return delete(new File(path));
    }

    /**
     * @param f file/directory to be deleted
     * @return true if deletion was OK
     */
    public static boolean delete(final File f) {
        if (f.isFile()) {
            return f.delete();
        }
        // else (if directory)
        File[] children = f.listFiles();
        if (children == null) {
            // no children
            return f.delete();
        }

        // There are children
        boolean result = true;
        for (int i = 0; i < children.length; i++) {
            result &= delete(children[i]);
        }
        return result && f.delete();
    }

    /**
     * Copy a directory recursively.
     * The destination directory will be overridden.
     * @param src source directory
     * @param dest dest directory
     * @throws FileUtilsException if the copy of the directory failed
     */
    public static void copyDirectory(final String src, final String dest) throws FileUtilsException {
        copyDirectory(new File(src), new File(dest), true);
    }

    /**
     * Copy a directory recursively.
     * @param src source directory
     * @param dest dest directory
     * @param override true if the destination directory is overridden
     * @return the copied directory
     * @throws FileUtilsException if the copy of the directory failed
     */
    public static File copyDirectory(final File src, final File dest, final boolean override) throws FileUtilsException {

        if (!src.isDirectory()) {
            // We do not accept file arguments !
            throw new IllegalArgumentException("Source '" + src + "' must be a directory");
        }

        final File newDest;
        if (!dest.exists() || override) {
            newDest = dest;
        } else {
            // append source directory filename to the destination directory.
            newDest = new File(dest, src.getName());
        }

        // create the destination directory if it is inexistent
        if (!newDest.exists()) {
            newDest.mkdirs();
        }

        // copy the files of the source directory
        File[] childs = src.listFiles();
        if (childs != null) {
            // children
            for (int i = 0; i < childs.length; i++) {
                File child = childs[i];
                if (child.isFile()) {
                    // file
                    copyFile(child, newDest);
                } else {
                    // directory
                    copyDirectory(child, new File(newDest, child.getName()), true);
                }
            }
        }

        return newDest;
    }

    /**
     * Copy a directory recursively.
     * The destination directory will be overridden.
     * @param src source directory
     * @param dest dest directory
     * @throws FileUtilsException if the copy of the directory failed
     */
    public static void copyDirectory(final File src, final File dest) throws FileUtilsException {
        copyDirectory(src, dest, true);
    }

    /**
     * Change the attribute value in a jar file (and update file if required).
     * @param jarName output jar file
     * @param attributes a list of attributes (name <-> value)
     * @throws FileUtilsException if the manifest is not updated.
     */
    public static void updateAttributesInManifest(final String jarName,
                                                  final Map<String, String> attributes)
            throws FileUtilsException {

        // Check if the given File is really here
        File file = new File(jarName);
        if (!file.exists()) {
            throw new FileUtilsException("The file '" + file + "' is not existing on the filesystem");
        }

        // The processing is different if file or directory
        if (file.isFile()) {
            // file
            updateManifestAttributesInJarArchive(file, attributes);
        } else {
            // directory
            updateManifestAttributesInJarDirectory(file, attributes);
        }
    }

    /**
     * Change the attribute value in a jar file (and update directory if required).
     * @param file input archive
     * @param attributes a list of attributes (name <-> value)
     * @throws FileUtilsException if the given file is not a directory (it may be an archive)
     *                            or if the inner Manifest opened/updated.
     */
    public static void updateManifestAttributesInJarDirectory(final File file,
                                                              final Map<String, String> attributes)
            throws FileUtilsException {

        // First check if this is really a directory
        if (!file.isDirectory()) {
            throw new FileUtilsException("'" + file + "' is not a jar file directory.");
        }

        File manifestFile = new File(new File(file, "META-INF"), "MANIFEST.MF");
        Manifest manifest = new Manifest();

        if (!manifestFile.isFile()) {
            // The Manifest file probably not exists
            // Ensure that there is the MANIFEST_VERSION
            Attributes mainAttributes = manifest.getMainAttributes();
            mainAttributes.putValue("Manifest-Version", "1.0");
        } else {
            // Read the manifest from the existing file
            InputStream is = null;
            try {
                is = new FileInputStream(manifestFile);
                manifest.read(is);
            } catch (IOException e) {
                throw new FileUtilsException("Cannot read META-INF/MANIFEST.MF file for unpacked archive '"
                                             + file + "'", e);
            } finally {
                close(is);
                is = null;
            }
        }

        // Update values (don't bother if we changed something or not)
        setManifestMainAttributes(manifest, attributes);

        // Write it back
        OutputStream os = null;
        try {
            // Ensure that parents directory exists
            File parent = manifestFile.getParentFile();
            if (!parent.exists()) {
                parent.mkdirs();
            }
            os = new FileOutputStream(manifestFile);
            manifest.write(os);
        } catch (IOException e) {
            throw new FileUtilsException("Cannot write META-INF/MANIFEST.MF file for unpacked archive '"
                                         + file + "'", e);
        } finally {
            close(os);
            os = null;
        }
    }

    /**
     * Change the attribute value in a jar file (and update file if required).
     * @param file input archive
     * @param attributes a list of attributes (name <-> value)
     * @throws FileUtilsException if the given file is not a file (it may be a directory)
     *                            or if the given jar cannot be opened/updated/closed.
     */
    public static void updateManifestAttributesInJarArchive(final File file,
                                                            final Map<String, String> attributes)
            throws FileUtilsException {

        // First check if this is really a file
        if (!file.isFile()) {
            throw new FileUtilsException("'" + file + "' is not a jar file archive.");
        }

        // Regular file case
        // Create jarFile on it and read manifest
        JarFile jarFile = null;
        Manifest manifest = null;
        try {
            jarFile = new JarFile(file);
            manifest = jarFile.getManifest();
        } catch (IOException e) {
            throw new FileUtilsException("Error while using jar file '" + file + "'", e);
        } finally {
            if (jarFile != null) {
                try {
                    jarFile.close();
                } catch (IOException e) {
                    throw new FileUtilsException("Error while closing jar file '" + jarFile + "'", e);
                }
            }
        }

        // Check if manifest is up-to-date. If not, update it and change a flag
        boolean manifestUpdated = false;

        if (manifest == null) {
            manifest = new Manifest();
            // Ensure that there is the MANIFEST_VERSION
            Attributes mainAttributes = manifest.getMainAttributes();
            mainAttributes.putValue("Manifest-Version", "1.0");
            manifestUpdated = true;
        }
        manifestUpdated |= setManifestMainAttributes(manifest, attributes);

        // manifest has not been updated, values are up-to-date ? then stop
        // everything
        if (!manifestUpdated) {
            return;
        }

        // Temp file
        File tmpJarFile = null;
        try {
            tmpJarFile = File.createTempFile("tmpGenICFile", ".jar");
            updateJarFile(file, manifest, tmpJarFile);
            FileUtils.copyFile(tmpJarFile, file);
        } catch (IOException e) {
            throw new FileUtilsException("Cannot create temporary file", e);
        } finally {
            tmpJarFile.delete();
        }
    }

    /**
     * Close all given Closeable instances. Ignore IOException that may be thrown.
     * @param closeables list of Closeable instance to be closed.
     */
    public static void close(final Closeable... closeables) {
        if (closeables != null) {
            for (Closeable closeable : closeables) {
                if (closeable != null) {
                    try {
                        closeable.close();
                    } catch (IOException e) {
                        // Ignored
                        logger.debug("Cannot close the given Closeable", e);
                    }
                }
            }
        }
    }

    /**
     * Flush all given Flushable instances. Ignore IOException that may be thrown.
     * @param flushables list of Flushable instance to be closed.
     */
    public static void flush(final Flushable... flushables) {
        if (flushables != null) {
            for (Flushable flushable : flushables) {
                if (flushable != null) {
                    try {
                        flushable.flush();
                    } catch (IOException e) {
                        // Ignored
                        logger.debug("Cannot flush the given Flushable", e);
                    }
                }
            }
        }
    }

    /**
     * Change the permissions of a given file by using chmod.
     * @param file a file
     * @param mode a mode
     * @return true if chmod has been executed successfully
     */
    public static final boolean chmod(final File file, final String mode) {

        //TODO Don't use process in Java7 thanks to nio2
        final String osName = System.getProperty("os.name");
        if (osName != null && !osName.startsWith("Windows")) {
            final Process process;
            try {
                process = Runtime.getRuntime().exec(
                        new String[] {"chmod", mode, file.getAbsolutePath()});
            } catch (IOException e) {
                logger.debug(
                        "I/O error when executing chmod {0} {1}",
                        mode, file.getAbsolutePath(), e);
                return false;
            }
            while(true) {
                try {
                    return process.waitFor() == 0;
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
            }
        } else {
            logger.debug("Cannot execute chmod on the OS: {0}", osName);
            return false;
        }
    }

    public static final Collection<File> find(
            final File directory, final FileFilter fileFilter, final boolean recursive) {

        //TODO Use visitors in Java7 thanks to nio2
        if (directory.isDirectory()) {
            ArrayList<File> files = new ArrayList<File>();
            for (File file : directory.listFiles()) {
                if (file.isDirectory()) {
                    if (recursive) {
                        files.addAll(find(file,fileFilter,true));
                    }
                } else if (file.isFile() && fileFilter.accept(file)) {
                    files.add(file);
                }
            }
            return files;
        } else {
            throw new IllegalArgumentException("Parameter 'directory' is not a directory");
        }
    }

    /**
     * Return the mode for a given file.
     * @param file a file
     * @return a mode
     */
    public static final String getMode(final File file) {

        final String osName = System.getProperty("os.name");
        if (osName != null && !osName.startsWith("Windows")) {
            final Process process;
            try {
                process = Runtime.getRuntime().exec(
                        new String[] {"stat", "-c", "%a", file.getAbsolutePath()});
            } catch (IOException e) {
                logger.debug(
                        "I/O error when executing stat -c %a",
                        file.getAbsolutePath(), e);
                return null;
            }

            final StringBuilder sb = new StringBuilder();

            final Lock lock = new ReentrantLock();
            final Condition condition = lock.newCondition();

            final Runnable output = new Runnable() {

                public void run() {
                    final InputStreamReader isr =
                        new InputStreamReader(process.getInputStream());
                    final BufferedReader br = new BufferedReader(isr);

                    lock.lock();
                    try {
                        String line = null;
                        while ((line = br.readLine()) != null) {
                            sb.append(line);
                        }
                        condition.signal();
                    } catch (IOException e) {
                        logger.debug("Error when reading a new line", e);
                    } finally {
                        FileUtils.close(br, isr);
                        lock.unlock();
                    }
                }
            };

            lock.lock();
            try {
                new Thread(output).start();

                try {
                    condition.await();
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
            } finally {
                lock.unlock();
            }

            return sb.toString();
        } else {
            logger.debug("Cannot execute stat on the OS: {0}", osName);
            return null;
        }
    }

    /**
     * Update the Manifest Main attributes using the provided attributes Map.
     * @param manifest Manifest to be updated
     * @param attributes injected attributes
     * @return <code>true</code> if the Manifest was changed, <code>false</code> otherwise.
     */
    private static boolean setManifestMainAttributes(final Manifest manifest,
                                                     final Map<String, String> attributes) {

        boolean manifestUpdated = false;
        // Check given attributes with manifest attributes
        Attributes manifestAttributes = manifest.getMainAttributes();

        Set<Map.Entry<String, String>> entries = attributes.entrySet();
        for (Map.Entry<String, String> entry : entries) {
            String attributeName = entry.getKey();
            String attributeValue = entry.getValue();

            // exist in manifest
            String manifestAttributeValue = manifestAttributes.getValue(attributeName);
            if (attributeValue != null && !attributeValue.equals(manifestAttributeValue)) {
                // Update manifest
                Attributes.Name manifestAttributeName = new Attributes.Name(attributeName);
                manifestAttributes.put(manifestAttributeName, attributeValue);
                manifestUpdated = true;
            }
        }
        return manifestUpdated;
    }

    /**
     * For a given file, get the shorter name containing the last modification
     * date.
     * @param file the given file to analyze to get the file name and the last
     *        modification date
     * @return A new archive name which contains the last modification date of
     *         the given deployable
     * @throws FileUtilsException Thrown if the given file cannot be retrieved
     */
    public static String lastModifiedFileName(final File file) throws FileUtilsException {
        if (!file.exists()) {
            throw new FileUtilsException("Cannot get File '" + file + "'.");
        }

        // Get last modification date
        Date lastModificationDate = new Date(file.lastModified());

        // Create a format
        SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT);

        return file.getName().replaceAll("\\..{3}$", sdf.format(lastModificationDate) + "$0");
    }

    /**
     * For a given file, calculate its checksum using a given algorithm.
     * @param file The file to check
     * @param algorithm The checksum algorithm
     * @return The checksum of a given file.
     * @throws FileUtilsException If the checksum cannot be calculated
     */
    public static String calculateChecksum(final File file, final String algorithm) throws FileUtilsException {
        MessageDigest messageDigest = null;
        try {
            messageDigest = MessageDigest.getInstance(algorithm);
        } catch (NoSuchAlgorithmException e) {
            throw new FileUtilsException("Failed to create digester for the " + algorithm + " algorithm.");
        }

        InputStream inputStream = null;
        try {
            inputStream = new FileInputStream(file);
            byte[] buffer = new byte[1024];
            int read;
            while ((read = inputStream.read(buffer)) != -1) {
                if (read > 0) {
                    messageDigest.update(buffer, 0, read);
                }
            }
        } catch (IOException e) {
            throw new FileUtilsException("Cannot read file " + file + ".");
        } finally {
            close(inputStream);
        }

        byte[] bytes = messageDigest.digest();
        StringBuffer result = new StringBuffer();
        for (int i = 0; i < bytes.length; i++) {
            result.append(Integer.toString((bytes[i] & 0xff) + 0x100, 16).substring(1));
        }
        return result.toString();
    }

    /**
     * Update the given file with the given manifest and write result in
     * tmpJarFile.
     * @param file the jarfile which have the content
     * @param manifest the manifest to use
     * @param tmpJarFile the resulting file
     * @throws FileUtilsException if jar file is not updated
     */
    private static void updateJarFile(final File file,
                                      final Manifest manifest,
                                      final File tmpJarFile) throws FileUtilsException {

        // File Stream
        FileOutputStream fileOS = null;
        try {
            fileOS = new FileOutputStream(tmpJarFile);
        } catch (FileNotFoundException e) {
            throw new FileUtilsException("Cannot create file output stream on the file '" + file + "'", e);
        }

        // Creates an ouputStream with the given manifest
        JarOutputStream fileJarOutputStream = null;
        try {
            fileJarOutputStream = new JarOutputStream(fileOS, manifest);
        } catch (IOException e) {
            throw new FileUtilsException("Cannot create Jar OutputStream", e);
        }

        JarFile jarFile = null;
        try {

            int n;
            byte[] buffer = new byte[BUFFER_SIZE];
            jarFile = new JarFile(file);
            Enumeration<JarEntry> jarEntries = jarFile.entries();

            while (jarEntries.hasMoreElements()) {

                // Get Entry
                JarEntry jarEntry = jarEntries.nextElement();

                // Ignore Manifest entry as it is already added by
                // JarOutputStream
                if (JarFile.MANIFEST_NAME.equals(jarEntry.getName())) {
                    continue;
                }

                // Store entry
                fileJarOutputStream.putNextEntry(new JarEntry(jarEntry.getName()));

                // Put input stream
                InputStream inputStream = jarFile.getInputStream(jarEntry);
                try {

                    n = inputStream.read(buffer);
                    while (n > 0) {
                        fileJarOutputStream.write(buffer, 0, n);
                        n = inputStream.read(buffer);
                    }
                    fileJarOutputStream.closeEntry();
                    // fileJarOutputStream.flush();
                } finally {
                    if (inputStream != null) {
                        inputStream.close();
                    }
                }

            }
        } catch (IOException ioe) {
            throw new FileUtilsException("Cannot update jar file '" + file + "'", ioe);
        } finally {
            // Close streams
            try {
                fileJarOutputStream.flush();
                fileJarOutputStream.close();
            } catch (IOException e) {
                throw new FileUtilsException("Cannot close jar file output stream", e);
            }

            try {
                fileOS.close();
            } catch (IOException e) {
                throw new FileUtilsException("Cannot close file output stream", e);
            }

            try {
                if (jarFile != null) {
                    jarFile.close();
                }
            } catch (IOException e) {
                throw new FileUtilsException("Cannot close jar file '" + jarFile + "'.", e);
            }
        }
    }

    /**
     * Return the string contained in the given file, by using the given charset to decode it.
     * @param path the file path
     * @param charset the charset to decode the file
     * @return the string contained in the given file
     * @throws IOException if some I/O errors occur
     */
    public static String readFile(final String path, final Charset charset) throws IOException {

        final FileInputStream stream = new FileInputStream(new File(path));
        try {
            final FileChannel fc = stream.getChannel();
            final MappedByteBuffer bb = fc.map(FileChannel.MapMode.READ_ONLY, 0, fc.size());
            return charset.decode(bb).toString();
        } finally {
            FileUtils.close(stream);
        }
    }

    /**
     * Return the string contained in the given file, by using the default charset.
     * @param path the file path
     * @return the string contained in the given file
     * @throws IOException if some I/O errors occur
     */
    public static String readFile(final String path) throws IOException {

        return readFile(path, Charset.defaultCharset());
    }


}
